@@ -3,10 +3,14 @@ from collections import Counter, Sequence, Iterable, \
     Mapping
 from functools import partial
 import gc
+from email.mime.image import MIMEImage
+from email.mime.multipart import MIMEMultipart
 from email.mime.text import MIMEText
+from email import encoders
 import inspect
 from itertools import chain
 import json
+from multiprocessing import Pool
 import os
 from pathlib import Path
 import pickle
@@ -14,6 +18,7 @@ from random import choice
 import re
 import smtplib
 import sys
+from tqdm.auto import tqdm
 
 from htools.config import get_credentials, get_default_user
 
@@ -115,7 +120,8 @@ def hasarg(func, arg):
     return arg in inspect.signature(func).parameters
 
 
-def quickmail(subject, message, to_email, from_email=None):
+def quickmail(subject, message, to_email, from_email=None, img_path=None,
+              img_name=None, verbose=True):
     """Send an email.
 
     Parameters
@@ -133,28 +139,35 @@ def quickmail(subject, message, to_email, from_email=None):
     --------
     None
     """
-    # Load source email address.
+    # Load email username. Error handling takes place in config functions.
     from_email = from_email or get_default_user()
-    if not from_email:
-        return None
+    if not from_email: return None
 
     # Load email password.
     password = get_credentials(from_email)
-    if not password:
-        return None
+    if not password: return None
 
-    # Create message instance.
-    msg = MIMEText(message)
+    # Create message and add text if specified.
+    msg = MIMEMultipart()
     msg['Subject'] = subject
     msg['From'] = from_email
     msg['To'] = to_email
+    if message: msg.attach(MIMEText(message))
+
+    # Load and attach image.
+    if img_path:
+        with open(img_path, 'rb') as f:
+            img = MIMEImage(f.read(),
+                            name=img_name or os.path.basename(img_path))
+        encoders.encode_base64(img)
+        msg.attach(img)
 
     # Access server and send email.
     server = smtplib.SMTP(host='smtp.gmail.com', port=587)
     server.starttls()
     server.login(user=from_email, password=password)
-    server.send_message(msg)
-    print(f'Email sent to {to_email}.')
+    server.sendmail(from_email, to_email, msg.as_string())
+    if verbose: print(f'Email sent to {to_email}.')
 
 
 def hsplit(text, sep, group=True, attach=True):
@@ -884,6 +897,38 @@ def smap(*x):
     return amap('shape', *x)
 
 
+def parallelize(func, items, total=None, chunksize=1_000, processes=None):
+    """Apply a function to a sequence of items in parallel. A progress bar
+    is included.
+
+    Parameters
+    ----------
+    func: function
+        This will be applied to each item in `items`.
+    items: Iterable
+        Sequence of items to apply `func` to.
+    total: int or None
+        This defaults to the length of `items`. In the case that items is a
+        generator, this lets us pass in the length explicitly. This lets tdqm
+        know how quickly to advance our progress bar.
+    chunksize: int
+        Positive int that determines the size of chunks submitted to the
+        process pool as separate tasks. Multiprocessing's default is 1 but
+        larger values should speed things up, especially with long sequences.
+    processes: None
+        Optionally set number of processes to run in parallel.
+
+    Returns
+    -------
+
+    """
+    total = total or len(items)
+    with Pool(processes) as p:
+        res = list(tqdm(p.imap(func, items, chunksize=chunksize),
+                        total=total))
+    return res
+
+
 def identity(x):
     """Returns the input argument. Sometimes it is convenient to have this if
     we sometimes apply a function to an item: rather than defining a None
@@ -902,6 +947,14 @@ def identity(x):
     return x
 
 
+def always_true(x, *args, **kwargs):
+    """Similar to `identity` but returns True instead of x. I'm tempted to name
+    this `true` but I fear that will cause some horrible bugs where I
+    accidentally use this when I want to use True.
+    """
+    return True
+
+
 def ifnone(arg, backup):
     """Shortcut to provide a backup value if an argument is None. Commonly used
     for numpy arrays since their truthiness is ambiguous.
@@ -920,6 +973,50 @@ def ifnone(arg, backup):
     return arg if arg is not None else backup
 
 
+def listlike(x):
+    """Checks if an object is a list/tuple/set/array etc. Strings and
+    mappings (e.g. dicts) are not considered list-like.
+    """
+    return isinstance(x, Iterable) and not isinstance(x, (str, Mapping))
+
+
+def tolist(x, strict=False):
+    """Helper to let a function accept a single value or a list of values for
+    a certain parameter. See examples.
+
+    Parameters
+    ----------
+    x: Iterable
+        Usually an object that could either be a list/tuple or a primitive,
+        depending on what user passed in.
+    strict: bool
+        If True, returned value will always be a list. If False, we allow
+        tuples/sets/etc. to retain their initial type.
+
+    Returns
+    -------
+    Iterable: list if strict is True or if x is a primitive. If strict is
+    False and x is already a tuple/set/something similar, its type will be
+    retained.
+
+    Examples
+    --------
+    def train(lrs):
+        lrs = tolist(lrs)
+        ...
+
+    >>> train(3e-3)
+    >>> train([3e-4, 3e-3])
+    """
+    if isinstance(x, Mapping):
+        raise ValueError('x must not be a mapping. It should probably be a '
+                         'primitive (str, int, etc.) or a list-like object '
+                         '(tuple, list, set).')
+    elif isinstance(x, Iterable) and not isinstance(x, str):
+        return list(x) if strict else x
+    return [x]
+
+
 def xor_none(*args, n=1):
     """Checks that exactly 1 (or n) of inputs is not None. Useful for
     validating optional function arguments (for example, ensuring the user
@@ -1190,7 +1287,9 @@ def kwargs_fallback(self, *args, assign=False, **kwargs):
     """
     res = []
     for arg in args:
-        val = kwargs.get(arg) or getattr(self, arg)
+        # Don't just use `kwargs.get(arg) or ...` because this doesn't work
+        # well when we pass in a numpy array or None.
+        val = kwargs[arg] if arg in kwargs else getattr(self, arg)
         res.append(val)
         if assign: setattr(self, arg, val)
     return res if len(res) > 1 else res[0]
